/*
* Copyright c                  Realtek Semiconductor Corporation, 2005
* All rights reserved.
* 
* Program :	rtl8651_dns_domainBlock.c
* Abstract :	DNS based domain blocking
* Creator :	Yi-Lun Chen (chenyl@realtek.com.tw)
* Author :	Yi-Lun Chen (chenyl@realtek.com.tw)
*
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_errno.h"
#include "rtl_glue.h"
#include "mbuf.h"

#include "mbuf.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_dns_local.h"
#include "rtl8651_dns_domainBlock_local.h"

static _rtl8651_domainBlockingBlockList_t *_rtl8651_domainBlocking_findBlockList(_rtl8651_domainBlockingInfo_t *info, ipaddr_t mapped_ip);
static int32 _rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlockingInfo_t *info, uint8 fromDbModule);
static int32 _rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlockingBlockList_t *ptr, uint8 refreshACL);



/* ==========================================================================
  *
  *		DNS Domain blocking:
  *
  *			Block special DNS domain by ROMEDRV.
  *
  *			1. Get the IP information about Blocked Domain by DNS request / reply.
  *			2. Block the specific IP address of these domain.
  *			3. System need to provide a mechanism to free the blocked IP address.
  *
  *
  * ========================================================================== */
static uint8							_rtl8651_domainBlocking_procSequence		= 0;
static uint32							 _rtl8651_domainBlocking_blockListTimeout	= 0;
static _rtl8651_domainBlockingInfo_t		*_rtl8651_domainBlocking_info_inused		= NULL;
static _rtl8651_domainBlockingInfo_t		*_rtl8651_domainBlocking_info_freeList		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_h		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_t		= NULL;
static _rtl8651_domainBlockingBlockList_t	*_rtl8651_domainBlocking_blockList_freeList	= NULL;

/*
	Procedure to register Domain Block module into DNS module.
*/
int32 _rtl8651_registerDomainBlock(void)
{
	return _rtl8651_dns_registerProcess(	&_rtl8651_domainBlocking_procSequence,
										RTL8651_DNSDB_PROPERTY_KEYWORDMATCH,
										_rtl8651_domainBlocking_init,
										_rtl8651_domainBlocking_flushBlockList,
										_rtl8651_domainBlocking_procMapping,
										_rtl8651_domainBlocking_timeUpdate,
										_rtl8651_domainBlocking_blockInfoDestructor);
}

/* =========================================================================
  *
  *
  *
  *
  *										Call back functions
  *
  *
  *
  *
  * ========================================================================= */

#define RTL8651_DOMAINBLOCK_CALLBACK

/* INIT function */
int32 _rtl8651_domainBlocking_init(void *_para)
{
	rtl8651_fwdEngineInitPara_t *para = (rtl8651_fwdEngineInitPara_t *)_para;

	_rtl8651_domainBlocking_info_inused = NULL;
	_rtl8651_domainBlocking_info_freeList = NULL;
	_rtl8651_domainBlocking_blockList_h = NULL;
	_rtl8651_domainBlocking_blockList_t = NULL;
	_rtl8651_domainBlocking_blockList_freeList = NULL;
	_rtl8651_domainBlocking_blockListTimeout = 0;

	if (	(para->domainBlock_tblSize == 0) ||
		(para->domainBlock_maxBlockEntryCnt == 0) ||
		(para->domainBlock_blockEntryTimeout == 0))
	{
		RTL_DNS_INFO("Disable [domain blocking]:\n\tdomainBlock_tblSize (%d) domainBlock_maxBlockEntryCnt (%d) domainBlock_blockEntryTimeout (%d)\n",
			para->domainBlock_tblSize,
			para->domainBlock_maxBlockEntryCnt,
			para->domainBlock_blockEntryTimeout);
		return SUCCESS;
	}

	MEM_ALLOC(_rtl8651_domainBlocking_info_freeList, sizeof(_rtl8651_domainBlockingInfo_t) * para->domainBlock_tblSize, FAILED);
	DNS_LIST_INIT(_rtl8651_domainBlocking_info_freeList, _rtl8651_domainBlockingInfo_t, para->domainBlock_tblSize, next);

	MEM_ALLOC(_rtl8651_domainBlocking_blockList_freeList, sizeof(_rtl8651_domainBlockingBlockList_t) * para->domainBlock_maxBlockEntryCnt, FAILED);
	DNS_LIST_INIT(_rtl8651_domainBlocking_blockList_freeList, _rtl8651_domainBlockingBlockList_t, para->domainBlock_maxBlockEntryCnt, next);

	_rtl8651_domainBlocking_blockListTimeout = para->domainBlock_blockEntryTimeout;

	return SUCCESS;	

}

/* REINIT function : we only flush block list here. Information block would be flushed by DNS module (using callback function 'destructor') */
void _rtl8651_domainBlocking_flushBlockList(void)
{
	uint8 refreshAcl = FALSE;

	while (_rtl8651_domainBlocking_blockList_h)
	{
		refreshAcl = TRUE;
		_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
}

/* PROC function */
int32 _rtl8651_domainBlocking_procMapping(
	char *orig_domain,
	char *mapped_domain,
	ipaddr_t mapped_ip,
	void *_info)
{
	_rtl8651_domainBlockingInfo_t *info = (_rtl8651_domainBlockingInfo_t *)_info;
	_rtl8651_domainBlockingBlockList_t *blkList;

	/* check if any block list has apply this mapping */
	if ((blkList = _rtl8651_domainBlocking_findBlockList(info, mapped_ip)) != NULL)
	{
		/* simply refresh it */
		blkList->updateTime = _sysUpSeconds;
		HTLIST_MOVE2TAIL(blkList, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
	} else
	{
		/* add new block for this mapping */
		if (_rtl8651_domainBlocking_blockList_freeList == NULL)
		{
			if (_rtl8651_domainBlocking_blockList_h)
			{
				_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
			}
		}
		
		if ((blkList = _rtl8651_domainBlocking_blockList_freeList) != NULL)
		{
			_rtl8651_domainBlocking_blockList_freeList = blkList->next;

			/* init entry */
			blkList->next = NULL;
			blkList->prev = NULL;
			blkList->info = info;
			blkList->updateTime = _sysUpSeconds;
			blkList->dip = mapped_ip;

			HTLIST_ADD2TAIL(blkList, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
		}
	}

	_rtl8651_arrangeAllACL();
	
	return SUCCESS;
}

/* TIMEUPDATE function */
void _rtl8651_domainBlocking_timeUpdate(uint32 secPassed)
{
	uint8 refreshAcl = FALSE;

	while (	_rtl8651_domainBlocking_blockList_h &&
			((_rtl8651_domainBlocking_blockList_h->updateTime + _rtl8651_domainBlocking_blockListTimeout) < _sysUpSeconds))
	{
		refreshAcl = TRUE;
		_rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlocking_blockList_h, FALSE);
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
}


/* DESTRUCTOR */
int32 _rtl8651_domainBlocking_blockInfoDestructor(void *info)
{
	return _rtl8651_domainBlocking_delBlockInfo((_rtl8651_domainBlockingInfo_t *)info, FALSE);
}


/* =========================================================================
  *
  *
  *										External functions.
  *
  *			1. for other functions in DRIVER	- without MUTEX protection
  *			2. for USERs					- with MUTEX protection
  *
  *
  * ========================================================================= */
#define RTL8651_DOMAINBLOCK_EXTERNAL

/*
@func int32										| rtl8651_domainBlocking_addDomainBlockingEntry | Add domain blocking entry
@parm char*						| domain_name	| Domain name keyword to block.
@parm rtl8651_domainBlockingInfo_t	| usrInfo			| Related user informations to further block traffic.
@rvalue SUCCESS					| Entry addition success.
@rvalue FAILED					| Entry addition failed.
@comm
	Add domain blocking entry into database to block any traffic sent to the target domain.
	This is keyword match, so when user define 'abc' in <p domain_name> , all domain names with the keyword 'abc' would hit..
	By the way, this domain blocking function provide <p usrInfo> to allow user set more constraint to filter traffic.
	Now, it allows user to set sip/sipMask pair to filter traffic which from only sip/sipMask.
*/
int32 rtl8651_domainBlocking_addDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	int32 retval;

	if (!domain_name)
		return FAILED;

	rtlglue_drvMutexLock();
	retval = _rtl8651_domainBlocking_addDomainBlockingEntry(domain_name, usrInfo);
	rtlglue_drvMutexUnlock();

	return retval;
}

int32 _rtl8651_domainBlocking_addDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	void *tmp = NULL;
	char *domain;
	rtl8651_domainBlockingInfo_t dummy, *usedUsrInfo = usrInfo;
	_rtl8651_domainBlockingInfo_t *info;

	if (domain_name == NULL)
	{
		return FAILED;
	}

	if (usrInfo == NULL)
	{
		memset(&dummy, 0, sizeof(dummy));
		usedUsrInfo = &dummy;
	}
	/* check redundant setting */
	while (_rtl8651_dns_findProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											NULL,
											(char**)&domain,
											(void*)&info,
											(void*)&tmp) == TRUE)
	{
		assert(info);
		if (	(strcmp(domain, domain_name) == 0) &&
			(memcmp(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t)) == 0))
		{	/* redunduncy found */
			return SUCCESS;
		}
	}

	/* add new block info */
	if (_rtl8651_domainBlocking_info_freeList)
	{
		info = _rtl8651_domainBlocking_info_freeList;
		if (_rtl8651_dns_addProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											(void*)info,
											&(info->ctrlEntry)) == SUCCESS)
		{
			HLIST_REMOVE(info, next, prev, _rtl8651_domainBlocking_info_freeList);
			HLIST_ADD2HDR(info, next, prev, _rtl8651_domainBlocking_info_inused);
			rtl8651_memcpy(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t));
			return SUCCESS;
		}
	}

	return FAILED;
}

/*
@func int32										| rtl8651_domainBlocking_delDomainBlockingEntry | Delete domain blocking entry
@parm char*						| domain_name	| Domain name keyword to block.
@parm rtl8651_domainBlockingInfo_t	| usrInfo			| Related user informations to further block traffic.
@rvalue SUCCESS					| Entry deletion success (might entry is not found!).
@rvalue FAILED					| Entry deletion failed.
@comm
	Delete domain blocking entry from database.
	Note that user MUST set same information (both <p domain_name> and <p usrInfo> ) to delete entry.
*/

int32 rtl8651_domainBlocking_delDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	int32 retval;

	if (!domain_name)
		return FAILED;

	rtlglue_drvMutexLock();
	retval = _rtl8651_domainBlocking_delDomainBlockingEntry(domain_name, usrInfo);
	rtlglue_drvMutexUnlock();

	return retval;
}


int32 _rtl8651_domainBlocking_delDomainBlockingEntry(char *domain_name, rtl8651_domainBlockingInfo_t *usrInfo)
{
	void *tmp = NULL;
	char *domain;
	rtl8651_domainBlockingInfo_t dummy, *usedUsrInfo = usrInfo;
	_rtl8651_domainBlockingInfo_t *info;

	if (domain_name == NULL)
	{
		return FAILED;
	}

	if (usrInfo == NULL)
	{
		memset(&dummy, 0, sizeof(dummy));
		usedUsrInfo = &dummy;
	}

	/* find entry to delete */
	while (_rtl8651_dns_findProcessEntry(	_rtl8651_domainBlocking_procSequence,
											domain_name,
											NULL,
											(char**)&domain,
											(void*)&info,
											(void*)&tmp) == TRUE)
	{
		assert(info);
		if (	(strcmp(domain, domain_name) == 0) &&
			(memcmp(&(info->info), usedUsrInfo, sizeof(rtl8651_domainBlockingInfo_t)) == 0))
		{	/* entry found */
			if (_rtl8651_domainBlocking_delBlockInfo(info, TRUE) != SUCCESS)
			{
				return FAILED;
			}
			/* we need to refresh "tmp" for entry search */
			tmp = NULL;
		}
	}
	return SUCCESS;
}

/*
@func int32			| rtl8651_domainBlocking_flushDomainBlockingEntry | Flush all domain blocking entry
@rvalue SUCCESS		| Entry flush success.
@rvalue FAILED		| Entry flush failed.
@comm
	Flush all domain blocking entries in database.
*/
int32 rtl8651_domainBlocking_flushDomainBlockingEntry(void)
{
	int32 ret;

	rtlglue_drvMutexLock();
	ret = _rtl8651_domainBlocking_flushDomainBlockingEntry();
	rtlglue_drvMutexUnlock();

	return ret;
}

int32 _rtl8651_domainBlocking_flushDomainBlockingEntry(void)
{
	while( _rtl8651_domainBlocking_info_inused )
	{
		if (_rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlocking_info_inused, TRUE) != SUCCESS)
		{
			return FAILED;
		}
	}
	return SUCCESS;
}

/*
@func void			| rtl8651_domainBlocking_dump | Dump domain blocking system's information
@comm
	Dump information in domain blocking system.
*/
void rtl8651_domainBlocking_dump(void)
{
	uint32 inusedCnt = 0, freeCnt = 0;
	_rtl8651_domainBlockingInfo_t *info;
	_rtl8651_domainBlockingBlockList_t *blkList;

	rtlglue_drvMutexLock();

	info = _rtl8651_domainBlocking_info_inused;
	while (info)
	{
		inusedCnt ++;
		info = info->next;
	}
	info = _rtl8651_domainBlocking_info_freeList;
	while (info)
	{
		freeCnt ++;
		info = info->next;
	}

	printk("\nDOMAIN BLOCK table:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);

	if (inusedCnt)
	{
		int32 idx;
		char *domain;

		printk("\n\t[NO] [SIP/MASK]\t\t\t\t\t[PTR]\n");
		printk("\t============================================================\n");
		info = _rtl8651_domainBlocking_info_inused;
		idx = 1;
		while (info)
		{
			domain = _rtl8651_dns_getProcessEntryDomainInfo(info->ctrlEntry);
			printk("\t%2d    %03u.%03u.%03u.%03u/%03u.%03u.%03u.%03u\t\t%p\n\t\t\t=> (%s)\n",
				idx,
				NIPQUAD(info->info.sip),
				NIPQUAD(info->info.sipMask),
				info,
				domain?domain:"(NULL)");
			idx ++;
			info = info->next;
		}
		printk("\t============================================================\n");
	}

	inusedCnt = 0;
	freeCnt = 0;

	blkList = _rtl8651_domainBlocking_blockList_h;
	while (blkList)
	{
		inusedCnt ++;
		blkList = blkList->next;
	}
	blkList = _rtl8651_domainBlocking_blockList_freeList;
	while (blkList)
	{
		freeCnt ++;
		blkList = blkList->next;
	}

	printk("\nBlocking list:\n\n");
	printk("\tInused:\t\t%d\n", inusedCnt);
	printk("\tFree:\t\t%d\n", freeCnt);

	if (inusedCnt)
	{
		char *domain;

		printk("\n\t[TOUT]\t[SIP/MASK]\t\t\t[DIP]\n");
		printk("\t============================================================\n");
		blkList = _rtl8651_domainBlocking_blockList_h;
		while (blkList)
		{
			info = blkList->info;
			domain = _rtl8651_dns_getProcessEntryDomainInfo(info->ctrlEntry);
			printk("\t%5d  %03u.%03u.%03u.%03u/%03u.%03u.%03u.%03u\t(%03u.%03u.%03u.%03u)\n\t\t\t=> (%s)\n",
				blkList->updateTime + _rtl8651_domainBlocking_blockListTimeout - _sysUpSeconds,
				NIPQUAD(info->info.sip),
				NIPQUAD(info->info.sipMask),
				NIPQUAD(blkList->dip),
				domain?domain:"(NULL)");
			blkList = blkList->next;
		}
		printk("\t============================================================\n");
	}

	rtlglue_drvMutexUnlock();
}

/*

	Add DROP-ACL-rule to drop packet in blocking list.

*/
#define SET_DMBLKACL_MATCHLAYER(rule) \
	rule.pktOpApp = RTL8651_ACLTBL_L3_AND_L4;

uint32 _rtl8651_tryToAddAclRuleForDnsDomainBlocking(rtl8651_tblDrv_vlanTable_t *vlanPtr, uint32 aclStart, uint32 aclEnd, int8 *enoughFlag)
{
	_rtl8651_tblDrvAclRule_t rule;
	uint32 addCount = 0, aclIdx = aclStart;

	if ( vlanPtr->internal )	/* only add blocking rule to internal vlan */
	{
		_rtl8651_domainBlockingBlockList_t *blkList = _rtl8651_domainBlocking_blockList_h;

		while ( blkList )
		{
			if ( aclIdx > aclEnd )
			{
				*enoughFlag = 0;	/* Not enough space for adding IPSec ACL rules. */
				goto out;
			}

			bzero(&rule, sizeof(_rtl8651_tblDrvAclRule_t));

			rule.ruleType_			= RTL8651_ACL_IP;
			rule.actionType_		= RTL8651_ACL_DROP;
			rule.srcIpAddr_		= blkList->info->info.sip;
			rule.srcIpAddrMask_	= blkList->info->info.sipMask;
			rule.dstIpAddr_		= blkList->dip;
			rule.dstIpAddrMask_	= 0xffffffff;

			SET_DMBLKACL_MATCHLAYER(rule);
			rtl8651_setAsicAclRule(aclIdx, &rule);

			addCount ++;
			aclIdx ++;
			blkList = blkList->next;
		}
	}

out:
	return addCount;
}


/* =========================================================================
  *
  *
  *										Internal used functions.
  *
  *
  * ========================================================================= */
#define RTL8651_DOMAINBLOCK_INTERNAL

/*
	Find blocked list entry in domain blocking system.
*/
static _rtl8651_domainBlockingBlockList_t *_rtl8651_domainBlocking_findBlockList(_rtl8651_domainBlockingInfo_t *info, ipaddr_t mapped_ip)
{
	_rtl8651_domainBlockingBlockList_t *ptr;

	ptr = _rtl8651_domainBlocking_blockList_h;

	while (ptr)
	{
		if ((ptr->info == info) && (ptr->dip == mapped_ip))
		{
			return ptr;
		}
		ptr = ptr->next;
	}

	return NULL;
}

/*
	Delete block information entry in domain blocking system.
*/
static int32 _rtl8651_domainBlocking_delBlockInfo(_rtl8651_domainBlockingInfo_t *info, uint8 fromDbModule)
{
	_rtl8651_domainBlockingBlockList_t * ptr, *prev;
	uint8 refreshAcl = FALSE;

	if (fromDbModule == TRUE)
	{
		/*
		  *	This function is called by Domain Blocking module,
		  *	we remove (Domain Blocking info entry) using DNS module's "destructor" mechanism.
		  *	So we only call function to remove DNS module entry here.
		  *
		  */

		/* remove corresponding control entry in DNS module */
		if (_rtl8651_dns_delProcessEntry(	_rtl8651_domainBlocking_procSequence,
										TRUE,
										info->ctrlEntry,
										NULL,
										NULL) != SUCCESS)
		{
			return FAILED;
		}
	} else
	{
		/*
		  *	This function is called by DNS module, to destroy domain blocking info entry.
		  *	We just set the info entry back to free list.
		  *
		  */

		/* remove from inused list */
		HLIST_REMOVE(info, next, prev, _rtl8651_domainBlocking_info_inused);
		HLIST_ADD2HDR(info, next, prev, _rtl8651_domainBlocking_info_freeList);
	}

	/* trace block list to delete corresponding entries */
	prev = NULL;
	ptr = _rtl8651_domainBlocking_blockList_h;
	while (ptr)
	{
		if (ptr->info == info)
		{
			_rtl8651_domainBlocking_delBlockList(ptr, FALSE);
			refreshAcl = TRUE;
			ptr = prev;
		} else
		{
			prev = ptr;
			ptr = ptr->next;
		}
	}

	if (refreshAcl == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}

	return SUCCESS;

}

/*
	Delete blocking list entry in domain blocking system.
*/
static int32 _rtl8651_domainBlocking_delBlockList(_rtl8651_domainBlockingBlockList_t *ptr, uint8 refreshACL)
{
	if (ptr == NULL)
	{
		return FAILED;
	}
	/* remove entry from LRU list */
	HTLIST_REMOVE(ptr, next, prev, _rtl8651_domainBlocking_blockList_h, _rtl8651_domainBlocking_blockList_t);
	HLIST_ADD2HDR(ptr, next, prev, _rtl8651_domainBlocking_blockList_freeList);
	if (refreshACL == TRUE)
	{
		_rtl8651_arrangeAllACL();
	}
	return SUCCESS;
}



