/*
**
** File:    lpc.c
**
** Description: Functions that implement linear predictive coding 
**      (LPC) operations.  
**
** Functions:
**
**  Computing LPC coefficients:
**
**      Comp_Lpc()
**      Durbin()
**
**  Perceptual noise weighting:
**
**      Wght_Lpc()
**      Error_Wght()
**
**  Computing combined impulse response:
**
**      Comp_Ir()
**
**  Computing ringing response:
**
**      Sub_Ring()
**      Upd_Ring()
**
**  Synthesizing speech:
**
**      Synt()
**      Spf()
*/

/*
    ITU-T G.723 Speech Coder   ANSI-C Source Code     Version 5.00
    copyright (c) 1995, AudioCodes, DSP Group, France Telecom,
    Universite de Sherbrooke.  All rights reserved.
*/

//#include <stdio.h>

#include "../basic_op/typedef.h"
#include "../basic_op/fixedpt_basic_op.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lbccodec.h"
#include "coder.h"
#include "decod.h"
#include "util_lbc.h"
#include "lpc.h"
#include "cod_cng.h"

void Fir(Word32 *Acc0,	//(i/o) 1st sample accumulator
		 Word32 *Acc2,	//(i/o) 2nd sample accumulator
		 Word16 *pCoe,	//(i)
		 Word16 *pDat)	//(i)
{
	*Acc0 = L_msu( *Acc0, pCoe[0], pDat[0]); *Acc0 = L_msu( *Acc0, pCoe[1], pDat[1]);
	*Acc0 = L_msu( *Acc0, pCoe[2], pDat[2]); *Acc0 = L_msu( *Acc0, pCoe[3], pDat[3]);
	*Acc0 = L_msu( *Acc0, pCoe[4], pDat[4]); *Acc0 = L_msu( *Acc0, pCoe[5], pDat[5]);
	*Acc0 = L_msu( *Acc0, pCoe[6], pDat[6]); *Acc0 = L_msu( *Acc0, pCoe[7], pDat[7]);
	*Acc0 = L_msu( *Acc0, pCoe[8], pDat[8]); *Acc0 = L_msu( *Acc0, pCoe[9], pDat[9]);
        
	*Acc2 = L_msu( *Acc2, pCoe[0], pDat[-1]); *Acc2 = L_msu( *Acc2, pCoe[1], pDat[0]);
	*Acc2 = L_msu( *Acc2, pCoe[2], pDat[1]); *Acc2 = L_msu( *Acc2, pCoe[3], pDat[2]);
	*Acc2 = L_msu( *Acc2, pCoe[4], pDat[3]); *Acc2 = L_msu( *Acc2, pCoe[5], pDat[4]);
	*Acc2 = L_msu( *Acc2, pCoe[6], pDat[5]); *Acc2 = L_msu( *Acc2, pCoe[7], pDat[6]);
	*Acc2 = L_msu( *Acc2, pCoe[8], pDat[7]); *Acc2 = L_msu( *Acc2, pCoe[9], pDat[8]);
}

void Iir(Word32 *Acc0,	// (i/o) first sample accumulator
		 Word32 *Acc2,	// (i/o) 2nd sample accumulator
		 Word16 *pCoe,	// (i) coeff
		 Word16 *pDat)	// (i) data
						// processing 2 samples each time
{
	Word32	Acc1, Acc3;
	
	*Acc0 = L_mac(*Acc0, pCoe[0], pDat[0]);	*Acc0 = L_mac(*Acc0, pCoe[1], pDat[1]);
	*Acc0 = L_mac(*Acc0, pCoe[2], pDat[2]);	*Acc0 = L_mac(*Acc0, pCoe[3], pDat[3]);
	*Acc0 = L_mac(*Acc0, pCoe[4], pDat[4]);	*Acc0 = L_mac(*Acc0, pCoe[5], pDat[5]);
	*Acc0 = L_mac(*Acc0, pCoe[6], pDat[6]);	*Acc0 = L_mac(*Acc0, pCoe[7], pDat[7]);
	*Acc0 = L_mac(*Acc0, pCoe[8], pDat[8]);	*Acc0 = L_mac(*Acc0, pCoe[9], pDat[9]);        

											*Acc2 = L_mac(*Acc2, pCoe[1], pDat[0]);
	*Acc2 = L_mac(*Acc2, pCoe[2], pDat[1]);	*Acc2 = L_mac(*Acc2, pCoe[3], pDat[2]);
	*Acc2 = L_mac(*Acc2, pCoe[4], pDat[3]);	*Acc2 = L_mac(*Acc2, pCoe[5], pDat[4]);
	*Acc2 = L_mac(*Acc2, pCoe[6], pDat[5]);	*Acc2 = L_mac(*Acc2, pCoe[7], pDat[6]);
	*Acc2 = L_mac(*Acc2, pCoe[8], pDat[7]);	*Acc2 = L_mac(*Acc2, pCoe[9], pDat[8]);
	
	Acc1 = L_shl( *Acc0, (Word16) 2 ) ;
	pDat[-1] = round( Acc1 );

	*Acc2 = L_mac(*Acc2, pCoe[0], pDat[-1]);

	Acc3 = L_shl( *Acc2, (Word16) 2);
	pDat[-2] = round( Acc3 );

}

/*
	crossconv180()
	1. Len is multiple of 4
	2. Len[0]: number of output
	3. Len[1]: inner loop length
	4. integer mpy
*/
void crossconv180(Word16 *Tv, Word16 *Imr, Word32 *CrossEn, Word16 *Len)
{
	int		i,j;

	for ( i = 0 ; i < Len[0] ; i ++ ) {
        CrossEn[i] = (Word32) 0 ;
        for ( j = i ; j < Len[1] ; j ++ )            
			CrossEn[i] = L_add(CrossEn[i], Tv[j]*Imr[j-i]);
    }
}

/*
**
** Function:        Comp_Lpc()
**
** Description:     Computes the tenth-order LPC filters for an
**          entire frame.  For each subframe, a
**          Hamming-windowed block of 180 samples,
**          centered around the subframe, is used to
**          compute eleven autocorrelation coefficients.
**          The Levinson-Durbin algorithm then generates
**          the LPC coefficients.  This function requires
**          a look-ahead of one subframe, and hence
**          introduces a 7.5 ms encoding delay.
**
** Links to text:   Section 2.4
**
** Arguments:
**
**  Word16 *UnqLpc      Empty Buffer
**  Word16 PrevDat[]    Previous 2 subframes of samples (120 words)
**  Word16 DataBuff[]   Current frame of samples (240 words)
**
** Outputs:

**
**  Word16 UnqLpc[]     LPC coefficients for entire frame (40 words)
**
** Return value:    None
**
*/
void  Comp_Lpc( Word16 *UnqLpc, Word16 *PrevDat, Word16 *DataBuff )
{
    int   i,j,k ;

    Word16   Dpnt[Frame+LpcFrame-SubFrLen] ;
    //Word16   Vect[LpcFrame] ;
	Word16	 *Vect;
	Word16	 Vect_Norm[LpcFrame+4];				// cz 920523
    Word16   Acf_sf[LpcOrderP1*SubFrames];
    Word16   ShAcf_sf[SubFrames];	
    Word16   Exp   ;
    Word16   *curAcf;
    Word16   Pk2;

    Word32   Acc0,Acc1   ;
	Word32	 CrossEn[12];
	Word16	 Len[2];

    /*
     * Generate a buffer of 360 samples.  This consists of 120 samples
     * from the previous frame and 240 samples from the current frame.
     */
    for ( i = 0 ; i < LpcFrame-SubFrLen ; i ++ )
        Dpnt[i] = PrevDat[i] ;
    for ( i = 0 ; i < Frame ; i ++ )
        Dpnt[i+LpcFrame-SubFrLen] = DataBuff[i] ;


    /*
     * Repeat for all subframes
     */
    curAcf = Acf_sf;
    for ( k = 0 ; k < SubFrames ; k ++ ) {

        /*
        * Do windowing
        */

        /* Get block of 180 samples centered around current subframe */
        //for ( i = 0 ; i < LpcFrame ; i ++ )
        //    Vect[i] = Dpnt[k*SubFrLen+i] ;
		Vect = &Dpnt[k*SubFrLen];

        /* Normalize */
#ifdef VC6
        ShAcf_sf[k] = Vec_Norm( Vect, Vect_Norm, (Word16) LpcFrame ) ;
#else
		ShAcf_sf[k] = Vec_Norms( Vect, Vect_Norm, (Word16) LpcFrame ) ;
#endif

        /* Apply the Hamming window */
        for ( i = 0 ; i < LpcFrame ; i ++ )
            Vect_Norm[i] = mult_r(Vect_Norm[i], HammingWindowTable[i]) ;

		/* pad 4 zero in the tail of Vect_Norm */
		Vect_Norm[LpcFrame]=0;
		Vect_Norm[LpcFrame+1]=0;
		Vect_Norm[LpcFrame+2]=0;
		Vect_Norm[LpcFrame+3]=0;

        /*
        * Compute the autocorrelation coefficients
        */

        /* Compute the zeroth-order coefficient (energy) */
        Acc1 = (Word32) 0 ;
        for ( i = 0 ; i < LpcFrame ; i ++ ) {
            //Acc0 = L_mult( Vect_Norm[i], Vect_Norm[i] ) ;
            //Acc0 = L_shr( Acc0, (Word16) 1 ) ;			
            //Acc1 = L_add( Acc1, Acc0 ) ;
			Acc1 = L_add(Acc1, Vect_Norm[i]*Vect_Norm[i]);	// 920527
        }

        /* Apply a white noise correction factor of (1025/1024) */
        Acc0 = L_shr( Acc1, (Word16) RidgeFact ) ;
        Acc1 = L_add( Acc1, Acc0 ) ;

        /* Normalize the energy */
        Exp = norm_l( Acc1 ) ;		// Acc1 >= 0
        //Acc1 = L_shl( Acc1, Exp ) ;		
		Acc1 = Acc1 << Exp;

        curAcf[0] = round( Acc1 ) ;
        if(curAcf[0] == 0) {
            for ( i = 1 ; i <= LpcOrder ; i ++ )
                curAcf[i] = 0;
            ShAcf_sf[k] = 40;
        }
        else {
            /* Compute the rest of the autocorrelation coefficients.
               Multiply them by a binomial coefficients lag window. */
			Len[0] = 12;
			Len[1] = LpcFrame;
			crossconv180(Vect_Norm, Vect_Norm, CrossEn, Len);
			for (i=1; i<=LpcOrder; i++)
			{
				Acc0 = CrossEn[i] << Exp;
				Acc0 = L_mls( Acc0, BinomialWindowTable[i-1] );
				curAcf[i] = round( Acc0 );
			}
			
			/*for ( i = 1 ; i <= LpcOrder ; i ++ ) {
                Acc1 = (Word32) 0 ;
                for ( j = i ; j < LpcFrame ; j ++ ) {
                    //Acc0 = L_mult( Vect_Norm[j], Vect_Norm[j-i] ) ;
                    //Acc0 = L_shr( Acc0, (Word16) 1 ) ;
                    //Acc1 = L_add( Acc1, Acc0 ) ;
					Acc1 = L_add(Acc1, Vect_Norm[j]*Vect_Norm[j-i]);	// 920527
                }
                //Acc0 = L_shl( Acc1, Exp ) ;
				Acc0 = Acc1 << Exp;
                Acc0 = L_mls( Acc0, BinomialWindowTable[i-1] ) ;
                curAcf[i] = round(Acc0) ;
            }*/
            /* Save Acf scaling factor */
//            ShAcf_sf[k] = add(Exp, shl(ShAcf_sf[k], 1));
			ShAcf_sf[k] = add(Exp, add(ShAcf_sf[k],ShAcf_sf[k]));
        }

        /*
         * Apply the Levinson-Durbin algorithm to generate the LPC
         * coefficients
        */
        Durbin( &UnqLpc[k*LpcOrder], &curAcf[1], curAcf[0], &Pk2 );
        pCoderMem->CodStat.SinDet <<= 1;
        if ( Pk2 > 0x799a ) {
            pCoderMem->CodStat.SinDet ++ ;
        }
        curAcf += LpcOrderP1;
    }

    /* Update sine detector */
    pCoderMem->CodStat.SinDet &= 0x7fff ;

    j = pCoderMem->CodStat.SinDet ;
    k = 0 ;
    for ( i = 0 ; i < 15 ; i ++ ) {
        k += j & 1 ;
        j >>= 1 ;
    }
    if ( k >= 14 )
        pCoderMem->CodStat.SinDet |= 0x8000 ;

    /* Update CNG Acf memories */
    Update_Acf(Acf_sf, ShAcf_sf);

}


/*
**
** Function:        Durbin()
**
** Description:     Implements the Levinson-Durbin algorithm for a
**          subframe.  The Levinson-Durbin algorithm
**          recursively computes the minimum mean-squared
**          error (MMSE) linear prediction filter based on the
**          estimated autocorrelation coefficients.
**
** Links to text:   Section 2.4
**
** Arguments:       
**
**  Word16 *Lpc Empty buffer
**  Word16 Corr[]   First- through tenth-order autocorrelations (10 words)
**  Word16 Err  Zeroth-order autocorrelation, or energy
**
** Outputs:     
**
**  Word16 Lpc[]    LPC coefficients (10 words)
**
** Return value:    The error
**
*/
Word16  Durbin( Word16 *Lpc, Word16 *Corr, Word16 Err, Word16 *Pk2 )
{
    int   i,j   ;

    Word16   Temp[LpcOrder] ;
    Word16   Pk ;

    Word32   Acc0,Acc1,Acc2 ;

 /*
  * Initialize the LPC vector
  */
    for ( i = 0 ; i < LpcOrder ; i ++ )
        Lpc[i] = (Word16) 0 ;

 /*
  * Do the recursion.  At the ith step, the algorithm computes the
  * (i+1)th - order MMSE linear prediction filter.
  */
    for ( i = 0 ; i < LpcOrder ; i ++ ) {

/*
 * Compute the partial correlation (parcor) coefficient
 */

        /* Start parcor computation */
        Acc0 = L_deposit_h( Corr[i] ) ;
        Acc0 = L_shr( Acc0, (Word16) 2 ) ;
        for ( j = 0 ; j < i ; j ++ )
            Acc0 = L_msu( Acc0, Lpc[j], Corr[i-j-1] ) ;
        Acc0 = L_shl( Acc0, (Word16) 2 ) ;

        /* Save sign */
        Acc1 = Acc0 ;
        Acc0 = L_abs( Acc0 ) ;

        /* Finish parcor computation */
        Acc2 = L_deposit_h( Err ) ;
        if ( Acc0 >= Acc2 ) {
            *Pk2 = 32767;
            break ;
        }

        Pk = div_l( Acc0, Err ) ;

        if ( Acc1 >= 0 )
            Pk = negate(Pk) ;

 /*
  * Sine detector
  */
        if ( i == 1 ) 
			*Pk2 = Pk;

 /*
  * Compute the ith LPC coefficient
  */
        Acc0 = L_deposit_h( negate(Pk) ) ;
        Acc0 = L_shr( Acc0, (Word16) 2 ) ;
        Lpc[i] = round( Acc0 ) ;

 /*
  * Update the prediction error
  */
        Acc1 = L_mls( Acc1, Pk ) ;
        Acc1 = L_add( Acc1, Acc2 ) ;
        Err = round( Acc1 ) ;

 /*
  * Compute the remaining LPC coefficients
  */
        for ( j = 0 ; j < i ; j ++ )
            Temp[j] = Lpc[j] ;

        for ( j = 0 ; j < i ; j ++ ) {
            Acc0 = L_deposit_h( Lpc[j] ) ;
            Acc0 = L_mac( Acc0, Pk, Temp[i-j-1] ) ;
            Lpc[j] = round( Acc0 ) ;
        }
    }

    return Err ;
}

/*
**
** Function:        Wght_Lpc()
**
** Description:     Computes the formant perceptual weighting
**          filter coefficients for a frame.  These
**          coefficients are geometrically scaled versions
**          of the unquantized LPC coefficients.
**
** Links to text:   Section 2.8  
**
** Arguments:       
**
**  Word16 *PerLpc      Empty Buffer
**  Word16 UnqLpc[]     Unquantized LPC coefficients (40 words)
**
** Outputs:     

**
**  Word16 PerLpc[]     Perceptual weighting filter coefficients
**              (80 words)
**
** Return value:    None
**
*/
void  Wght_Lpc( Word16 *PerLpc, Word16 *UnqLpc )
{
    int   i,j   ;


 /*
  * Do for all subframes
  */
    for ( i = 0 ; i < SubFrames ; i ++ ) {

		for (j=0; j<LpcOrder; j++) {
			PerLpc[j] = mult_r(UnqLpc[j], PerFiltZeroTable[j]);
			PerLpc[LpcOrder+j] = mult_r(UnqLpc[j], PerFiltPoleTable[j]);
		}
		PerLpc += 2*LpcOrder;
		UnqLpc += LpcOrder;

#if 0
 /*
  * Compute the jth FIR coefficient by multiplying the jth LPC
  * coefficient by (0.9)^j.
  */
        for ( j = 0 ; j < LpcOrder ; j ++ )
            PerLpc[j] = mult_r( UnqLpc[j], PerFiltZeroTable[j] ) ;
        PerLpc += LpcOrder ;


/*
 * Compute the jth IIR coefficient by multiplying the jth LPC
 * coefficient by (0.5)^j.
 */
        for ( j = 0 ; j < LpcOrder ; j ++ )
            PerLpc[j] = mult_r( UnqLpc[j], PerFiltPoleTable[j] ) ;
        PerLpc += LpcOrder ;
        UnqLpc += LpcOrder ;
#endif
    }
}

/*
**
** Function:        Error_Wght()
**
** Description:     Implements the formant perceptual weighting
**          filter for a frame. This filter effectively
**          deemphasizes the formant frequencies in the
**          error signal.
**
** Links to text:   Section 2.8
**
** Arguments:
**
**  Word16 Dpnt[]       Highpass filtered speech x[n] (240 words)
**  Word16 PerLpc[]     Filter coefficients (80 words)
**
** Inputs:
**
**  CodStat.WghtFirDl[] FIR filter memory from previous frame (10 words)
**  CodStat.WghtIirDl[] IIR filter memory from previous frame (10 words)

**
** Outputs:
**
**  Word16 Dpnt[]       Weighted speech f[n] (240 words)
**
** Return value:    None
**
*/
void  Error_Wght( Word16 *Dpnt, Word16 *PerLpc )
{
    int   i,k ;

    Word32   Acc0, Acc2  ;
	Word16	 FirDl[SubFrLen+LpcOrder], IirDl[SubFrLen+LpcOrder], *ptrF, *ptrI;


	/* Initial FirDl and IirDl */
		ptrF = &FirDl[SubFrLen];
		ptrI = &IirDl[SubFrLen];
		for(i=0;i<LpcOrder;i++)
		{
			ptrF[i] = pCoderMem->CodStat.WghtFirDl[i];
			ptrI[i] = pCoderMem->CodStat.WghtIirDl[i];
		}
/*
 * Do for all subframes
 */
    for ( k = 0 ; k < SubFrames ; k ++ ) {

        //for ( i = 0 ; i < SubFrLen ; i ++ ) {
		for(i=0;i<SubFrLen; i+=2) {

/*
 * Do the FIR part
 */
            /* Filter */
            Acc0 = L_mult( Dpnt[0], (Word16) 0x2000 ) ;
			Acc2 = L_mult( Dpnt[1], (Word16) 0x2000 ) ;

			ptrF[-1] = Dpnt[0];
			ptrF[-2] = Dpnt[1];

            //for ( j = 0 ; j < LpcOrder ; j ++ )
            //    Acc0 = L_msu( Acc0, PerLpc[j], ptrF[j]/*pCoderMem->CodStat.WghtFirDl[j]*/ ) ;
			Fir(&Acc0, &Acc2, PerLpc, ptrF);

            /* Update memory */
            /*for ( j = LpcOrder-1 ; j > 0 ; j -- )
                pCoderMem->CodStat.WghtFirDl[j] = pCoderMem->CodStat.WghtFirDl[j-1] ;
            pCoderMem->CodStat.WghtFirDl[0] = *Dpnt ;*/			

 /*
  * Do the IIR part
  */

            /* Filter */
            //for ( j = 0 ; j < LpcOrder ; j ++ )
            //    Acc0 = L_mac( Acc0, PerLpc[LpcOrder+j], ptrI[j]/*pCoderMem->CodStat.WghtIirDl[j]*/ ) ;
			Iir(&Acc0, &Acc2, &PerLpc[10], ptrI);            
            //Acc0 = L_shl( Acc0, (Word16) 2 ) ;
            /* Update memory */
			//ptrI[-1] = round( Acc0 );
            
			Dpnt[0] = ptrI[-1];
			Dpnt[1] = ptrI[-2];

			Dpnt += 2;
			ptrF -= 2;
			ptrI -= 2;
        }
        	
		/* restore memory */	
		for (i=0; i<LpcOrder; i++)
		{
			ptrF[SubFrLen+i] = ptrF[i];
			ptrI[SubFrLen+i] = ptrI[i];
        }
        PerLpc += 2*LpcOrder ;
		ptrF += SubFrLen;
		ptrI += SubFrLen;
    }

	// update memory	
	for (i=0; i<LpcOrder; i++)
	{
		pCoderMem->CodStat.WghtFirDl[i] = ptrF[i];
		pCoderMem->CodStat.WghtIirDl[i] = ptrI[i];
    }
}

/*
**
** Function:        Comp_Ir()
**
** Description:     Computes the combined impulse response of the
**          formant perceptual weighting filter, harmonic
**          noise shaping filter, and synthesis filter for
**          a subframe.
**
** Links to text:   Section 2.12
**
** Arguments:
**
**  Word16 *ImpResp     Empty Buffer
**  Word16 QntLpc[]     Quantized LPC coefficients (10 words)
**  Word16 PerLpc[]     Perceptual filter coefficients (20 words)
**  PWDEF Pw        Harmonic noise shaping filter parameters
**
** Outputs:
**
**  Word16 ImpResp[]    Combined impulse response (60 words)
**
** Return value:    None
**
*/

void  Comp_Ir( Word16 *ImpResp, Word16 *QntLpc, Word16 *PerLpc, PWDEF *Pw )
{
    int   i;

	Word16	 FirDl[SubFrLen+LpcOrder], IirDl[SubFrLen+LpcOrder], *ptrF, *ptrI;
    //Word16   Temp[PitchMax+SubFrLen] ;
	Word16   Temp[SubFrLen];
    Word32   Acc0,Acc2;
	Word16	 tmp0,tmp2, thr0;

 /*
  * Clear all memory.  Impulse response calculation requires
  * an all-zero initial state.
  */
	ptrF = &FirDl[SubFrLen];
	ptrI = &IirDl[SubFrLen];

    /* Perceptual weighting filter */
    for ( i = 0 ; i < LpcOrder ; i ++ )
		ptrF[i] = ptrI[i] = 0;

    /* Harmonic noise shaping filter */
    //for ( i = 0 ; i < PitchMax+SubFrLen ; i ++ )
    //    Temp[i] = (Word16) 0 ;


 /*
  * Input a single impulse
  */
    Acc0 = (Word32) 0x04000000L ;	Acc2 = 0;
	thr0 = -(Pw->Indx);

 /*
  * Do for all elements in a subframe
  */
    for ( i = 0 ; i < SubFrLen ; i += 2 ) {

 /*
  * Synthesis filter
  */
		Iir(&Acc0, &Acc2, QntLpc, ptrF);
		/*
		Acc0 = L_mac(Acc0, QntLpc[0], ptrF[0]);	Acc0 = L_mac(Acc0, QntLpc[1], ptrF[1]);
		Acc0 = L_mac(Acc0, QntLpc[2], ptrF[2]);	Acc0 = L_mac(Acc0, QntLpc[3], ptrF[3]);
		Acc0 = L_mac(Acc0, QntLpc[4], ptrF[4]);	Acc0 = L_mac(Acc0, QntLpc[5], ptrF[5]);
		Acc0 = L_mac(Acc0, QntLpc[6], ptrF[6]);	Acc0 = L_mac(Acc0, QntLpc[7], ptrF[7]);
		Acc0 = L_mac(Acc0, QntLpc[8], ptrF[8]);	Acc0 = L_mac(Acc0, QntLpc[9], ptrF[9]);        

												Acc2 = L_mac(Acc2, QntLpc[1], ptrF[0]);
		Acc2 = L_mac(Acc2, QntLpc[2], ptrF[1]);	Acc2 = L_mac(Acc2, QntLpc[3], ptrF[2]);
		Acc2 = L_mac(Acc2, QntLpc[4], ptrF[3]);	Acc2 = L_mac(Acc2, QntLpc[5], ptrF[4]);
		Acc2 = L_mac(Acc2, QntLpc[6], ptrF[5]);	Acc2 = L_mac(Acc2, QntLpc[7], ptrF[6]);
		Acc2 = L_mac(Acc2, QntLpc[8], ptrF[7]);	Acc2 = L_mac(Acc2, QntLpc[9], ptrF[8]);

        Acc1 = L_shl( Acc0, (Word16) 2 ) ;
		ptrF[-1] = round( Acc1 );

		Acc2 = L_mac(Acc2, QntLpc[0], ptrF[-1]);

		Acc3 = L_shl( Acc2, (Word16) 2);
		ptrF[-2] = round( Acc3 );
		*/
 /*
  * Perceptual weighting filter
  */
        /* FIR part */
		Fir(&Acc0, &Acc2, PerLpc, ptrF);
		/*
		Acc0 = L_msu( Acc0, PerLpc[0], ptrF[0]); Acc0 = L_msu( Acc0, PerLpc[1], ptrF[1]);
		Acc0 = L_msu( Acc0, PerLpc[2], ptrF[2]); Acc0 = L_msu( Acc0, PerLpc[3], ptrF[3]);
		Acc0 = L_msu( Acc0, PerLpc[4], ptrF[4]); Acc0 = L_msu( Acc0, PerLpc[5], ptrF[5]);
		Acc0 = L_msu( Acc0, PerLpc[6], ptrF[6]); Acc0 = L_msu( Acc0, PerLpc[7], ptrF[7]);
		Acc0 = L_msu( Acc0, PerLpc[8], ptrF[8]); Acc0 = L_msu( Acc0, PerLpc[9], ptrF[9]);
        
		Acc2 = L_msu( Acc2, PerLpc[0], ptrF[-1]); Acc2 = L_msu( Acc2, PerLpc[1], ptrF[0]);
		Acc2 = L_msu( Acc2, PerLpc[2], ptrF[1]); Acc2 = L_msu( Acc2, PerLpc[3], ptrF[2]);
		Acc2 = L_msu( Acc2, PerLpc[4], ptrF[3]); Acc2 = L_msu( Acc2, PerLpc[5], ptrF[4]);
		Acc2 = L_msu( Acc2, PerLpc[6], ptrF[5]); Acc2 = L_msu( Acc2, PerLpc[7], ptrF[6]);
		Acc2 = L_msu( Acc2, PerLpc[8], ptrF[7]); Acc2 = L_msu( Acc2, PerLpc[9], ptrF[8]);
		*/
                                Acc0 = L_shl( Acc0, (Word16) 1 ) ;
		Acc2 = L_shl( Acc2, (Word16) 1 );

        /* Iir part */
		Iir(&Acc0, &Acc2, &PerLpc[10], ptrI);
		/*
		Acc0 = L_mac( Acc0, PerLpc[10], ptrI[0]); Acc0 = L_mac( Acc0, PerLpc[11], ptrI[1]);
		Acc0 = L_mac( Acc0, PerLpc[12], ptrI[2]); Acc0 = L_mac( Acc0, PerLpc[13], ptrI[3]);
		Acc0 = L_mac( Acc0, PerLpc[14], ptrI[4]); Acc0 = L_mac( Acc0, PerLpc[15], ptrI[5]);
		Acc0 = L_mac( Acc0, PerLpc[16], ptrI[6]); Acc0 = L_mac( Acc0, PerLpc[17], ptrI[7]);
		Acc0 = L_mac( Acc0, PerLpc[18], ptrI[8]); Acc0 = L_mac( Acc0, PerLpc[19], ptrI[9]);
        

												  Acc2 = L_mac( Acc2, PerLpc[11], ptrI[0]);
		Acc2 = L_mac( Acc2, PerLpc[12], ptrI[1]); Acc2 = L_mac( Acc2, PerLpc[13], ptrI[2]);
		Acc2 = L_mac( Acc2, PerLpc[14], ptrI[3]); Acc2 = L_mac( Acc2, PerLpc[15], ptrI[4]);
		Acc2 = L_mac( Acc2, PerLpc[16], ptrI[5]); Acc2 = L_mac( Acc2, PerLpc[17], ptrI[6]);
		Acc2 = L_mac( Acc2, PerLpc[18], ptrI[7]); Acc2 = L_mac( Acc2, PerLpc[19], ptrI[8]);
        
        Acc0 = L_shl( Acc0, (Word16) 2 ) ;
		ptrI[-1] = round( Acc0 );

		Acc2 = L_mac( Acc2, PerLpc[10], ptrI[-1]);
		Acc2 = L_shl( Acc2, (Word16) 2 ) ;
		ptrI[-2] = round( Acc2 );
		*/

		//Temp[PitchMax+i] = ptrI[-1];
		//Temp[PitchMax+i+1] = ptrI[-2];
		Temp[i] = ptrI[-1];
		Temp[i+1] = ptrI[-2];

 /*
  * Harmonic noise shaping filter
  */
		//if (-Pw.Indx+i < 0)
		if (thr0 < 0)
			tmp0 = 0;
		else
			tmp0 = Temp[thr0];
		thr0++;

		if (thr0 < 0)
			tmp2 = 0;
		else
			tmp2 = Temp[thr0];
		thr0++;

		Acc0 = L_deposit_h( ptrI[-1] ) ;	Acc2 = L_deposit_h( ptrI[-2] ) ;	
        //Acc0 = L_msu( Acc0, Pw.Gain, Temp[PitchMax-Pw.Indx+i] ) ;	// Pw.Indx > 0		
		//Acc2 = L_msu( Acc2, Pw.Gain, Temp[PitchMax-Pw.Indx+i+1] ) ;
		Acc0 = L_msu( Acc0, Pw->Gain, tmp0);
		Acc2 = L_msu( Acc2, Pw->Gain, tmp2);
        ImpResp[i] = round( Acc0 ) ;		ImpResp[i+1] = round( Acc2 ) ;

        Acc0 = (Word32) 0 ;		Acc2 = (Word32) 0;

		ptrF -= 2;
		ptrI -= 2;
    }

}

/*
**
** Function:        Sub_Ring()
**
** Description:     Computes the zero-input response of the
**          combined formant perceptual weighting filter,
**          harmonic noise shaping filter, and synthesis
**          filter for a subframe.  Subtracts the
**          zero-input response from the harmonic noise
**          weighted speech vector to produce the target 
**          speech vector.
**
** Links to text:   Section 2.13
**
** Arguments:       
**
**  Word16 Dpnt[]       Harmonic noise weighted vector w[n] (60 words)
**  Word16 QntLpc[]     Quantized LPC coefficients (10 words)
**  Word16 PerLpc[]     Perceptual filter coefficients (20 words)
**  Word16 PrevErr[]    Harmonic noise shaping filter memory (145 words)
**  PWDEF Pw        Harmonic noise shaping filter parameters
**  
** Inputs:
**
**  CodStat.RingFirDl[] Perceptual weighting filter FIR memory from 
**               previous subframe (10 words)
**  CodStat.RingIirDl[] Perceptual weighting filter IIR memory from 
**               previous subframe (10 words)
**
** Outputs:     
**
**  Word16 Dpnt[]       Target vector t[n] (60 words)
**
** Return value:    None
**
*/
void  Sub_Ring( Word16 *Dpnt, Word16 *QntLpc, Word16 *PerLpc, Word16 *PrevErr, PWDEF *Pw )
{
    int   i;
    Word32   Acc0,Acc2;	
    //Word16   FirDl[LpcOrder] ;
    //Word16   IirDl[LpcOrder] ;
	Word16	 FirDl[SubFrLen+LpcOrder], IirDl[SubFrLen+LpcOrder], *ptrF, *ptrI;
    //Word16   Temp[PitchMax+SubFrLen] ;
	Word16	 Temp[SubFrLen], sam1, sam2, ind;
	
 /*
  * Initialize the memory
  */
    /*for ( i = 0 ; i < PitchMax ; i ++ )
        Temp[i] = PrevErr[i] ;*/
	PrevErr += PitchMax;

	ptrF = &FirDl[SubFrLen];
	ptrI = &IirDl[SubFrLen];
    for ( i = 0 ; i < LpcOrder ; i ++ ) {
        //FirDl[i] = CodStat.RingFirDl[i] ;
        //IirDl[i] = CodStat.RingIirDl[i] ;
		ptrF[i] = pCoderMem->CodStat.RingFirDl[i];
		ptrI[i] = pCoderMem->CodStat.RingIirDl[i];
    }

	ind = -((int)Pw->Indx);

 /*
  * Do for all elements in a subframe
  */
    //for ( i = 0 ; i < SubFrLen ; i ++ ) {
	for (i=0; i<SubFrLen; i+=2) {

 /*
  * Input zero
  */
        Acc0 = (Word32) 0 ;		Acc2 = 0;

 /*
  * Synthesis filter
  */
		Iir(&Acc0, &Acc2, QntLpc, ptrF);
        /*for ( j = 0 ; j < LpcOrder ; j ++ )
            Acc0 = L_mac( Acc0, QntLpc[j], FirDl[j] ) ;
        Acc1 = L_shl( Acc0, (Word16) 2 ) ;
		*/

 /*
  * Perceptual weighting filter
  */

        /* Fir part */
        //for ( j = 0 ; j < LpcOrder ; j ++ ) {
        //    Acc0 = L_msu( Acc0, PerLpc[j], ptrF[j] ) ;
	//		Acc2 = L_msu( Acc2, PerLpc[j], ptrF[-1+j] );
	//	}
        //for ( j = LpcOrder-1 ; j > 0 ; j -- )
        //    FirDl[j] = FirDl[j-1] ;
        //FirDl[0] = round( Acc1 ) ;		

		Fir(&Acc0, &Acc2, PerLpc, ptrF);


        /* Iir part */
		/*
        for ( j = 0 ; j < LpcOrder ; j ++ )
            Acc0 = L_mac( Acc0, PerLpc[LpcOrder+j], IirDl[j] ) ;
        Acc0 = L_shl( Acc0, (Word16) 2 ) ;
        //for ( j = LpcOrder-1 ; j > 0 ; j -- )
        //    IirDl[j] = IirDl[j-1] ;
        IirDl[0] = round( Acc0 ) ;
		*/

		Iir(&Acc0, &Acc2, &PerLpc[10], ptrI);
        //Temp[PitchMax+i] = IirDl[0] ;
		Temp[/*PitchMax+*/i] = ptrI[-1];
		Temp[/*PitchMax+*/i+1] = ptrI[-2];

 /*
  * Do the harmonic noise shaping filter and subtract the result
  * from the harmonic noise weighted vector.
  */
        //Acc0 = L_deposit_h( sub( Dpnt[i], IirDl[0] ) ) ;
		Acc0 = L_deposit_h( sub( Dpnt[i], ptrI[-1] ) ) ;
		Acc2 = L_deposit_h( sub( Dpnt[i+1], ptrI[-2] ) ) ;
				
		if ( ind < 0 )
			sam1 = PrevErr[ind];
		else
			sam1 = Temp[ind];
		ind++;
		if ( ind < 0 )
			sam2 = PrevErr[ind];
		else
			sam2 = Temp[ind];
		ind++;
		
        Acc0 = L_mac( Acc0, Pw->Gain, sam1/*Temp[PitchMax-(int)Pw->Indx+i]*/ ) ;
		Acc2 = L_mac( Acc2, Pw->Gain, sam2/*Temp[PitchMax-(int)Pw->Indx+i+1]*/ ) ;
        Dpnt[i] = round ( Acc0 ) ;
		Dpnt[i+1] = round ( Acc2 ) ;

		ptrF -= 2;
		ptrI -= 2;
    }
}

/*
**
** Function:        Upd_Ring()
**
** Description:     Updates the memory of the combined formant
**          perceptual weighting filter, harmonic noise
**          shaping filter, and synthesis filter for a
**          subframe.  The update is done by passing the
**          current subframe's excitation through the
**          combined filter.
**
** Links to text:   Section 2.19
**
** Arguments:       
**
**  Word16 Dpnt[]       Decoded excitation for the current subframe e[n] 
**               (60 words)
**  Word16 QntLpc[]     Quantized LPC coefficients (10 words)
**  Word16 PerLpc[]     Perceptual filter coefficients (20 words)
**  Word16 PrevErr[]    Harmonic noise shaping filter memory (145 words)
**  
** Inputs:
**
**  CodStat.RingFirDl[] Perceptual weighting filter FIR memory from 

**               previous subframe (10 words)
**  CodStat.RingIirDl[] Perceptual weighting filter IIR memory from 
**               previous subframe (10 words)
**
** Outputs:     
**
**  Word16 PrevErr[]    Updated harmonic noise shaping filter memory 
**  CodStat.RingFirDl[] Updated perceptual weighting filter FIR memory 
**  CodStat.RingIirDl[] Updated perceptual weighting filter IIR memory 
**
** Return value:    None
**
*/
void  Upd_Ring( Word16 *Dpnt, Word16 *QntLpc, Word16 *PerLpc, Word16 *PrevErr )
{
    int   i;
    Word32   Acc0;
	Word32	 Acc2;
	Word16	 FirDl[SubFrLen+LpcOrder], IirDl[SubFrLen+LpcOrder], *ptrF, *ptrI;
//	extern	 CODSTATDEF *pCodStat; 2005/04/20 Marked by BrianBin

	/*
		Initial FirDl, IirDl
	*/
	ptrF = &FirDl[SubFrLen];
	ptrI = &IirDl[SubFrLen];
	for (i=0; i<LpcOrder; i++)
	{
		ptrF[i] = pCoderMem->CodStat.RingFirDl[i];
		ptrI[i] = pCoderMem->CodStat.RingIirDl[i];		
	}

 /*
  * Shift the harmonic noise shaping filter memory
  */
/*    for ( i = SubFrLen ; i < PitchMax ; i ++ )
        PrevErr[i-SubFrLen] = PrevErr[i] ;
*/

 /*
  * Do for all elements in the subframe
  */
    //for ( i = 0 ; i < SubFrLen ; i ++ ) {
	for (i=0; i<SubFrLen; i+=2) {

 /*
  * Input the current subframe's excitation
  */
        Acc0 = L_deposit_h( Dpnt[i] ) ;
        Acc0 = L_shr( Acc0, (Word16) 3 ) ;

		Acc2 = L_deposit_h( Dpnt[i+1]);
		Acc2 = L_shr( Acc2, (Word16) 3 );

 /*
  * Synthesis filter
  */
//        for ( j = 0 ; j < LpcOrder ; j ++ ) {
//            Acc0 = L_mac( Acc0, QntLpc[j], ptrF[j]/*pCodStat->RingFirDl[j]*/ ) ;					
//		}
//        Acc1 = L_shl( Acc0, (Word16) 2 ) ;		
		Iir(&Acc0, &Acc2, QntLpc, ptrF);

		/*k = round(Acc1);		
		Dpnt[i] = add(k,k);*/
		Dpnt[i] = add(ptrF[-1], ptrF[-1]);
		Dpnt[i+1] = add(ptrF[-2], ptrF[-2]);

 /*
  * Perceptual weighting filter
  */

        /* FIR part */
        //for ( j = 0 ; j < LpcOrder ; j ++ )
        //    Acc0 = L_msu( Acc0, PerLpc[j], ptrF[j]/*pCodStat->RingFirDl[j]*/ ) ;

        /* Update FIR memory */
        /*for ( j = LpcOrder-1 ; j > 0 ; j -- )
            pCoderMem->CodStat.RingFirDl[j] = pCoderMem->CodStat.RingFirDl[j-1] ;
        pCoderMem->CodStat.RingFirDl[0] = round( Acc1 ) ;*/
		//ptrF[-1] = round( Acc1 );
		Fir(&Acc0, &Acc2, PerLpc, ptrF);

        /* IIR part */
        //for ( j = 0 ; j < LpcOrder ; j ++ )
        //    Acc0 = L_mac( Acc0, PerLpc[LpcOrder+j], ptrI[j]/*pCodStat->RingIirDl[j]*/ ) ;
        //Acc0 = L_shl( Acc0, (Word16) 2 ) ;
		Iir(&Acc0, &Acc2, &PerLpc[10], ptrI);

        /* Update IIR memory */
        /*for ( j = LpcOrder-1 ; j > 0 ; j -- )
            pCoderMem->CodStat.RingIirDl[j] = pCodStat->RingIirDl[j-1] ;
        pCoderMem->CodStat.RingIirDl[0] = round( Acc0 ) ;*/
		//ptrI[-1] = round( Acc0 );

        /* Update harmonic noise shaping memory */
        ////PrevErr[PitchMax-SubFrLen+i] = pCodStat->RingIirDl[0] ;
		//PrevErr[PitchMax+i] = /*pCodStat->RingIirDl[0]*/;
		PrevErr[PitchMax+i] = ptrI[-1];
		PrevErr[PitchMax+i+1] = ptrI[-2];

		ptrF -= 2;
		ptrI -= 2;
    }

	for (i=0; i<LpcOrder; i++)
	{
		pCoderMem->CodStat.RingFirDl[i] = ptrF[i];
		pCoderMem->CodStat.RingIirDl[i] = ptrI[i];
    }
}

/*
**
** Function:        Synt()
**
** Description:     Implements the decoder synthesis filter for a
**          subframe.  This is a tenth-order IIR filter.
**
** Links to text:   Section 3.7
**
** Arguments:       
**
**  Word16 Dpnt[]       Pitch-postfiltered excitation for the current
**               subframe ppf[n] (60 words)
**  Word16 Lpc[]        Quantized LPC coefficients (10 words)
**  
** Inputs:
**
**  DecStat.SyntIirDl[] Synthesis filter memory from previous
subframe (10 words)
**
** Outputs:     
**
**  Word16 Dpnt[]       Synthesized speech vector sy[n]
**  DecStat.SyntIirDl[] Updated synthesis filter memory 
**
** Return value:    None
**
*/
void     Synt( Word16 *Dpnt, Word16 *Lpc )
{
    int   i,j   ;

    Word32   Acc0, Acc1  ;
	Word16	 TmpVct[SubFrLen+LpcOrder], *pVct;		// 920609

	pVct = &TmpVct[SubFrLen];

	for (j=0; j<LpcOrder; j++)	
		pVct[j] = pDecodMem->DecStat.SyntIirDl[j];
	
	// pVct:
	// y(-1) y(-2) y(-3) ... y(-10)

 /*
  * Do for all elements in the subframe
  */
    //for ( i = 0 ; i < SubFrLen ; i ++ ) {
	for (i=0; i<SubFrLen; i+=2) {

 /*
  * Input the current subframe's excitation
  */
        Acc0 = L_deposit_h( Dpnt[i] ) ;
        Acc0 = L_shr( Acc0, (Word16) 3 ) ;
		// sample 2
		Acc1 = L_deposit_h( Dpnt[i+1] );
		Acc1 = L_shr( Acc1, (Word16) 3 );

 /*
  * Synthesis
  */

        /* Filter */
        //for ( j = 0 ; j < LpcOrder ; j ++ )
		/*for (j=0; j<LpcOrder; j+=2)
		{
            //Acc0 = L_mac( Acc0, Lpc[j], pDecStat->SyntIirDl[j] ) ;
			Acc0 = L_mac(Acc0, Lpc[j], pVct[j]);
			Acc0 = L_mac(Acc0, Lpc[j+1], pVct[j+1]);			
		}*/
		Iir(&Acc0, &Acc1, Lpc, pVct);
		/*
		Acc0 = L_mac(Acc0, Lpc[0], pVct[0]);
		Acc0 = L_mac(Acc0, Lpc[1], pVct[1]);
		Acc0 = L_mac(Acc0, Lpc[2], pVct[2]);
		Acc0 = L_mac(Acc0, Lpc[3], pVct[3]);
		Acc0 = L_mac(Acc0, Lpc[4], pVct[4]);
		Acc0 = L_mac(Acc0, Lpc[5], pVct[5]);
		Acc0 = L_mac(Acc0, Lpc[6], pVct[6]);
		Acc0 = L_mac(Acc0, Lpc[7], pVct[7]);
		Acc0 = L_mac(Acc0, Lpc[8], pVct[8]);
		Acc0 = L_mac(Acc0, Lpc[9], pVct[9]);

		Acc1 = L_mac(Acc1, 0, pVct[-1]);
		Acc1 = L_mac(Acc1, Lpc[1], pVct[0]);
		Acc1 = L_mac(Acc1, Lpc[2], pVct[1]);
		Acc1 = L_mac(Acc1, Lpc[3], pVct[2]);
		Acc1 = L_mac(Acc1, Lpc[4], pVct[3]);
		Acc1 = L_mac(Acc1, Lpc[5], pVct[4]);
		Acc1 = L_mac(Acc1, Lpc[6], pVct[5]);
		Acc1 = L_mac(Acc1, Lpc[7], pVct[6]);
		Acc1 = L_mac(Acc1, Lpc[8], pVct[7]);
		Acc1 = L_mac(Acc1, Lpc[9], pVct[8]);*/
		

        /* Update memory */
        ////for ( j = LpcOrder-1 ; j > 0 ; j -- )
        ////    pDecStat->SyntIirDl[j] = pDecStat->SyntIirDl[j-1] ;
		
        //Acc0 = L_shl( Acc0, (Word16) 2 ) ;

        ////pDecStat->SyntIirDl[0] = round( Acc0 ) ;
		//pVct[-1] = round(Acc0);		
		
		// sample 2
		//Acc1 = L_mac(Acc1, Lpc[0], pVct[-1]);
		//Acc1 = L_shl(Acc1, (Word16)2);
		//pVct[-2] = round(Acc1);

 /*
  * Scale output if postfilter is off.  (Otherwise output is
  * scaled by the gain scaling unit.)
  */
        if ( UsePf )
		{
            //Dpnt[i] = pDecStat->SyntIirDl[0] ;
			Dpnt[i] = pVct[-1];
			Dpnt[i+1] = pVct[-2];
		}
        else {
//            Dpnt[i] = shl( DecStat.SyntIirDl[0], (Word16) 1 ) ;
			//Dpnt[i] = add(pDecStat->SyntIirDl[0], pDecStat->SyntIirDl[0]);
			Dpnt[i] = add(pVct[-1], pVct[-1]);
			Dpnt[i+1] = add(pVct[-2], pVct[-2]);
		}

		pVct -= 2;
    }

	for (j=0; j<LpcOrder; j++)
		pDecodMem->DecStat.SyntIirDl[j] = pVct[j];


}

/*
**
** Function:        Spf()
**
** Description:     Implements the formant postfilter for a
**          subframe.  The formant postfilter is a
**          10-pole, 10-zero ARMA filter followed by a
**          single-tap tilt compensation filter.
**
** Links to text:   Section 3.8
**
** Arguments:
**
**  Word16 Tv[]     Synthesized speech vector sy[n] (60 words)
**  Word16 Lpc[]        Quantized LPC coefficients (10 words)
**
** Inputs:
**
**  DecStat.PostIirDl[] Postfilter IIR memory from previous
subframe (10 words)
**  DecStat.PostFirDl[] Postfilter FIR memory from previous
subframe (10 words)
**  DecStat.Park        Previous value of compensation filter parameter
**
** Outputs:
**
**  Word16 Tv[]     Postfiltered speech vector pf[n] (60 words)
**  DecStat.PostIirDl[] Updated postfilter IIR memory
**  DecStat.PostFirDl[] Updated postfilter FIR memory
**  DecStat.Park        Updated compensation filter parameter
**
** Return value: Input vector energy
**
*/
Word32  Spf( Word16 *Tv, Word16 *Lpc )
{
    int   i   ;

    Word32   Acc0,Acc1,L_tmp, L_tmp2   ;
    Word32   Sen ;
    Word16   Tmp ;
    Word16   Exp ;

    Word16   FirCoef[LpcOrder] ;
    Word16   IirCoef[LpcOrder] ;

    Word16   TmpVct[SubFrLen] ;

	Word16	 PastX[SubFrLen+LpcOrder], PastY[SubFrLen+LpcOrder], *pX, *pY;
    
 /*
  * Compute ARMA coefficients.  Compute the jth FIR coefficient by
  * multiplying the jth quantized LPC coefficient by (0.65)^j.
  * Compute the jth IIR coefficient by multiplying the jth quantized
  * LPC coefficient by (0.75)^j.  This emphasizes the formants in
  * the frequency response.
  */	
    for ( i = 0 ; i < LpcOrder ; i ++ ) {
        FirCoef[i] = mult_r( Lpc[i], PostFiltZeroTable[i] ) ;
        IirCoef[i] = mult_r( Lpc[i], PostFiltPoleTable[i] ) ;		
    }

 /*
  * Normalize the speech vector.
  */
#ifdef VC6  
	Exp = Vec_Norm( Tv, TmpVct, (Word16) SubFrLen ) ;
#else
	Exp = Vec_Norms( Tv, TmpVct, (Word16) SubFrLen ) ;
#endif

 /*
  * Compute the first two autocorrelation coefficients R[0] and R[1]
  */
    Acc0 = (Word32) 0 ;
    Acc1 = L_mult( TmpVct[0], TmpVct[0] ) ;
    for ( i = 1 ; i < SubFrLen ; i ++ ) {
        Acc0 = L_mac( Acc0, TmpVct[i], TmpVct[i-1] ) ;
        Acc1 = L_mac( Acc1, TmpVct[i], TmpVct[i] ) ;
    }

 /*
  * Scale the energy for the later use.
  */
//    Sen = L_shr( Acc1, (Word16)(2*Exp + 4) ) ;
	{
		Word16 fac = 2*Exp+4;

		if (fac >= 0)
			Sen = L_shr(Acc1, fac);
		else
			Sen = L_shl(Acc1, -fac);
	}

 /*
  * Compute the first-order partial correlation coefficient of the
  * input speech vector.
  */
    Tmp = extract_h( Acc1 ) ;
    if ( Tmp != (Word16) 0 ) {

        /* Compute first parkor */
        Acc0 = L_shr( Acc0, (Word16) 1 ) ;
        Acc1 = Acc0 ;
        Acc0 = L_abs( Acc0 ) ;

        Tmp = div_l( Acc0, Tmp ) ;

        if ( Acc1 < (Word32) 0 )
            Tmp = negate( Tmp ) ;
    }
    else
        Tmp = (Word16) 0 ;

 /*
  * Compute the compensation filter parameter and update the memory
  */
    Acc0 = L_deposit_h( pDecodMem->DecStat.Park ) ;
    Acc0 = L_msu( Acc0, pDecodMem->DecStat.Park, (Word16) 0x2000 ) ;
    Acc0 = L_mac( Acc0, Tmp, (Word16) 0x2000 ) ;
    pDecodMem->DecStat.Park = round( Acc0 ) ;

    Tmp  = mult( pDecodMem->DecStat.Park, PreCoef ) ;
    Tmp &= (Word16) 0xfffc ;

	// copy post memory to buffer
	pX = &PastX[SubFrLen];
	pY = &PastY[SubFrLen];
	for (i=0; i<LpcOrder; i++)
	{
		pX[i] = pDecodMem->DecStat.PostFirDl[i];
		pY[i] = pDecodMem->DecStat.PostIirDl[i];
	}

 /*
  *  Do for all elements in the subframe
  */
	
    //for ( i = 0 ; i < SubFrLen ; i ++ ) {
	for (i=0; i<SubFrLen; i+=2) {

 /*
  * Input the speech vector
  */		
        Acc0 = L_deposit_h( Tv[i] ) ;	
        Acc0 = L_shr( Acc0, (Word16) 2 ) ;

		Acc1 = L_deposit_h( Tv[i+1] );
		Acc1 = L_shr(Acc1, (Word16) 2);
		
 /*
  * Formant postfilter
  */    
		/*for (j=0; j<LpcOrder; j++)
		{
			Acc0 = L_msu(Acc0, FirCoef[j], pX[j]);		// Fir part
			Acc0 = L_mac(Acc0, IirCoef[j], pY[j]);		// Iir part			
		}*/
		pX[-1] = Tv[i];		
		pX[-2] = Tv[i+1];

		Fir(&Acc0, &Acc1, FirCoef, pX);
		Iir(&Acc0, &Acc1, IirCoef, pY);

		/*Acc0 = L_msu(Acc0, FirCoef[0], pX[0]);	Acc0 = L_msu(Acc0, FirCoef[1], pX[1]);		
		Acc0 = L_msu(Acc0, FirCoef[2], pX[2]);	Acc0 = L_msu(Acc0, FirCoef[3], pX[3]);
		Acc0 = L_msu(Acc0, FirCoef[4], pX[4]);	Acc0 = L_msu(Acc0, FirCoef[5], pX[5]);
		Acc0 = L_msu(Acc0, FirCoef[6], pX[6]);	Acc0 = L_msu(Acc0, FirCoef[7], pX[7]);
		Acc0 = L_msu(Acc0, FirCoef[8], pX[8]);	Acc0 = L_msu(Acc0, FirCoef[9], pX[9]);*/
		/*Acc0 = L_mac(Acc0, IirCoef[0], pY[0]);	Acc0 = L_mac(Acc0, IirCoef[1], pY[1]);
		Acc0 = L_mac(Acc0, IirCoef[2], pY[2]);	Acc0 = L_mac(Acc0, IirCoef[3], pY[3]);
		Acc0 = L_mac(Acc0, IirCoef[4], pY[4]);	Acc0 = L_mac(Acc0, IirCoef[5], pY[5]);
		Acc0 = L_mac(Acc0, IirCoef[6], pY[6]);	Acc0 = L_mac(Acc0, IirCoef[7], pY[7]);
		Acc0 = L_mac(Acc0, IirCoef[8], pY[8]);	Acc0 = L_mac(Acc0, IirCoef[9], pY[9]);*/

		/*Acc1 = L_msu(Acc1, FirCoef[0], pX[-1]);	Acc1 = L_msu(Acc1, FirCoef[1], pX[0]);
		Acc1 = L_msu(Acc1, FirCoef[2], pX[1]);	Acc1 = L_msu(Acc1, FirCoef[3], pX[2]);
		Acc1 = L_msu(Acc1, FirCoef[4], pX[3]);	Acc1 = L_msu(Acc1, FirCoef[5], pX[4]);
		Acc1 = L_msu(Acc1, FirCoef[6], pX[5]);	Acc1 = L_msu(Acc1, FirCoef[7], pX[6]);
		Acc1 = L_msu(Acc1, FirCoef[8], pX[7]);	Acc1 = L_msu(Acc1, FirCoef[9], pX[8]);*/
		/*Acc1 = L_mac(Acc1, 0, pY[-1]);			Acc1 = L_mac(Acc1, IirCoef[1], pY[0]);
		Acc1 = L_mac(Acc1, IirCoef[2], pY[1]);	Acc1 = L_mac(Acc1, IirCoef[3], pY[2]);
		Acc1 = L_mac(Acc1, IirCoef[4], pY[3]);	Acc1 = L_mac(Acc1, IirCoef[5], pY[4]);
		Acc1 = L_mac(Acc1, IirCoef[6], pY[5]);	Acc1 = L_mac(Acc1, IirCoef[7], pY[6]);
		Acc1 = L_mac(Acc1, IirCoef[8], pY[7]);	Acc1 = L_mac(Acc1, IirCoef[9], pY[8]);*/

        //Acc0 = L_shl( Acc0, (Word16) 2 ) ;
        L_tmp = L_shl(Acc0, 2) ;
		//pY[-1] = round(Acc0);
		
		//Acc1 = L_mac(Acc1, IirCoef[0], pY[-1]);
		//Acc1 = L_shl(Acc1, 2);
		L_tmp2 = L_shl(Acc1, 2);
		//pY[-2] = round(Acc1);

 /*
  * Compensation filter
  */        
		L_tmp = L_mac(L_tmp, pY[0], Tmp);
		L_tmp2 = L_mac(L_tmp2, pY[-1], Tmp);

        Tv[i] = round( L_tmp ) ;
		Tv[i+1] = round(L_tmp2);

		pX-=2;
		pY-=2;

    }

	// write back post memory
	for (i=0; i<LpcOrder; i++)
	{
		pDecodMem->DecStat.PostFirDl[i] = pX[i];
		pDecodMem->DecStat.PostIirDl[i] = pY[i];
	}

    return Sen ;
}

