/* 
 * leases.c -- tools to manage DHCP leases 
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 */

#include <time.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "dhcpd.h"
#include "files.h"
#include "options.h"
#include "arpping.h"
#include "common.h"
#include "leases.h"


unsigned char blank_chaddr[] = {[0 ... 15] = 0};

static char* value_parser_file(char *name)
{
    char *s,*sp;
    char buf[NVRAM_SIZE];
    int fd;
    unsigned int file_size;
    if((fd=open(NVRAM_TMP_PATH, O_RDONLY))<0)
    {
        //printf("Open file error!\n");
        return "";				
    }
    memset(buf,0,sizeof(buf));
    file_size = lseek(fd,0,SEEK_END);
    lseek(fd,0,SEEK_SET);
    //printf("file zise:%d",file_size);
    if (file_size>NVRAM_SIZE)
    {
        //printf("To value_parser_file function,the file size of nvram is big than the buf size!\n");
        close(fd);
        return "";
    }
    if(read(fd,buf,file_size)<0)
    {
        //printf("Read file nvram error!\n");
        close(fd);
        return "";
    }
    close(fd);
    s=buf;
    //printf("name:%s\n",name);
    //printf("name:%d\n",strlen(name));    
	while(*s) 
	{
		if (!strncmp(s, name, strlen(name)) && *(s+strlen(name))=='=') 
		{
			//sp=malloc(strlen(s)-strlen(name));
			//memcpy(sp,(s+strlen(name)+1),(strlen(s)-strlen(name)));						
			sp=s+strlen(name)+1;
			//printf("sp:%s\n",sp);
			return sp;
		}
		while(*s++);
	}
	return "";
}

/* clear every lease out that chaddr OR yiaddr matches and is nonzero */
void clear_lease(u_int8_t *chaddr, u_int32_t yiaddr)
{
	unsigned int i, j;
	
	for (j = 0; j < 16 && !chaddr[j]; j++);
	
	for (i = 0; i < server_config.max_leases; i++)
		if ((j != 16 && !memcmp(leases[i].chaddr, chaddr, 16)) ||
		    (yiaddr && leases[i].yiaddr == yiaddr)) {
			memset(&(leases[i]), 0, sizeof(struct dhcpOfferedAddr));
		}
}


/* add a lease into the table, clearing out any old ones */
struct dhcpOfferedAddr *add_lease(u_int8_t *chaddr, u_int32_t yiaddr, unsigned long lease,char *hostname)
{
	struct dhcpOfferedAddr *oldest;
	int size,i;

	if (hostname)
	{	 
		size=(int)(*(char*)(hostname-1));
	}
	/* clean out any old ones */
	clear_lease(chaddr, yiaddr);

	oldest = oldest_expired_lease();
	
	if (oldest) {
		memcpy(oldest->chaddr, chaddr, 16);
		oldest->yiaddr = yiaddr;
		oldest->expires = time(0) + lease;
		if (hostname)
		{
			if (size<sizeof(oldest->hostname))
				memcpy((char*)&oldest->hostname,(char*)hostname,size);
			else
				memcpy(oldest->hostname,hostname,sizeof(oldest->hostname));
		}

	}
	
	return oldest;
}


/* true if a lease has expired */
int lease_expired(struct dhcpOfferedAddr *lease)
{
	return (lease->expires < (unsigned long) time(0));
}	


/* Find the oldest expired lease, NULL if there are no expired leases */
struct dhcpOfferedAddr *oldest_expired_lease(void)
{
	struct dhcpOfferedAddr *oldest = NULL;
	unsigned long oldest_lease = time(0);
	unsigned int i;

	
	for (i = 0; i < server_config.max_leases; i++)
		if (oldest_lease > leases[i].expires) {
			oldest_lease = leases[i].expires;
			oldest = &(leases[i]);
		}
	return oldest;
		
}


struct dhcpManualAddr *find_manual_by_chaddr(u_int8_t *chaddr)
{
	unsigned int i;

	for(i=0; i < manualCount; i++)
		if (!memcmp(manualPool[i].chaddr, chaddr, 16)) return &(manualPool[i]);

	return NULL;
}


/* Find the first lease that matches chaddr, NULL if no match */
struct dhcpOfferedAddr *find_lease_by_chaddr(u_int8_t *chaddr)
{
	unsigned int i;

	for (i = 0; i < server_config.max_leases; i++)
		if (!memcmp(leases[i].chaddr, chaddr, 16)) return &(leases[i]);
	
	return NULL;
}


/* Find the first lease that matches yiaddr, NULL is no match */
struct dhcpOfferedAddr *find_lease_by_yiaddr(u_int32_t yiaddr)
{
	unsigned int i;

	for (i = 0; i < server_config.max_leases; i++)
		if (leases[i].yiaddr == yiaddr) return &(leases[i]);
	
	return NULL;
}


/* check is an IP is taken, if it is, add it to the lease table */
static int check_ip(u_int32_t addr)
{
	struct in_addr temp;

	if (arpping(addr, server_config.server, server_config.arp, server_config.interface) == 0) {
		temp.s_addr = addr;
		LOG(LOG_INFO, "%s belongs to someone, reserving it for %ld seconds",
			inet_ntoa(temp), server_config.conflict_time);
		add_lease(blank_chaddr, addr, server_config.conflict_time,0);
		return 1;
	} else return 0;
}

/* find an assignable address, it check_expired is true, we check all the expired leases as well.
 * Maybe this should try expired leases by age... */
u_int32_t find_address(int check_expired) 
{
	u_int32_t addr, ret;
	struct dhcpOfferedAddr *lease = NULL;		
    struct in_addr temp;
    char *lanip;
    
	addr = ntohl(server_config.start); /* addr is in host order here */
	for (;addr <= ntohl(server_config.end); addr++) {

		/* ie, 192.168.55.0 */
		if (!(addr & 0xFF)) continue;

		/* ie, 192.168.55.255 */
		if ((addr & 0xFF) == 0xFF) continue;
		temp.s_addr = addr;
        lanip = value_parser_file("lan_ipaddr");
       
		if (!strcmp(inet_ntoa(temp),lanip))
			continue;			
	
		/* lease is not taken */
		ret = htonl(addr);
		if ((!(lease = find_lease_by_yiaddr(ret)) ||

		     /* or it expired and we are checking for expired leases */
		     (check_expired  && lease_expired(lease))) &&

		     /* and it isn't on the network */
	    	     !check_ip(ret)) {
			return ret;
			break;
		}
	}
	return 0;
}

/* check whether the addr is in the lan ip address range or not
2 ...  same as the lan ip
1 ...  in the range
0 ...  not in the range */
int check_with_lanip(u_int32_t addr)
{
    u_int32_t lan_ip, lan_mask;
    
    lan_ip = inet_network(value_parser_file("lan_ipaddr"));
    lan_mask = inet_network(value_parser_file("lan_netmask"));
    
    //mBUG("in the check_with_lanip, addr =%x, lan_ip =%x, lan_mask =%x", addr, lan_ip, lan_mask);
    
    if(addr == lan_ip)
    {
        return 2;
    }
    else if((addr & lan_mask) == (lan_ip & lan_mask))
    {
        
        return 1;
    }
    else
    {
        return 0;
    }
}
    
